@@ -417,6 +417,36 @@ type_any_or_unknown(type_T *type)
 					       || type->tt_type == VAR_UNKNOWN;
 }
 
+/*
+ * Get a type_T for a partial typval in "tv".
+ */
+    static type_T *
+partial_typval2type(typval_T *tv, ufunc_T *ufunc, garray_T *type_gap)
+{
+    partial_T *pt = tv->vval.v_partial;
+    type_T  *type;
+
+    type = get_type_ptr(type_gap);
+    if (type == NULL)
+	return NULL;
+
+    *type = *ufunc->uf_func_type;
+    if (type->tt_argcount >= 0 && pt->pt_argc > 0)
+    {
+	type->tt_argcount -= pt->pt_argc;
+	type->tt_min_argcount -= pt->pt_argc;
+	if (type->tt_argcount > 0 && func_type_add_arg_types(type,
+		    type->tt_argcount, type_gap) == OK)
+	    for (int i = 0; i < type->tt_argcount; ++i)
+		type->tt_args[i] =
+		    ufunc->uf_func_type->tt_args[i + pt->pt_argc];
+    }
+    if (pt->pt_func != NULL)
+	type->tt_member = pt->pt_func->uf_ret_type;
+
+    return type;
+}
+
 /*
  * Get a type_T for a typval_T.
  * "type_gap" is used to temporarily create types in.
@@ -569,27 +599,8 @@ typval2type_int(typval_T *tv, int copyID, garray_T *type_gap, int flags)
 		set_function_type(ufunc);
 	    if (ufunc->uf_func_type != NULL)
 	    {
-		if (tv->v_type == VAR_PARTIAL && tv->vval.v_partial != NULL
-					    && tv->vval.v_partial->pt_argc > 0)
-		{
-		    type = get_type_ptr(type_gap);
-		    if (type == NULL)
-			return NULL;
-		    *type = *ufunc->uf_func_type;
-		    if (type->tt_argcount >= 0)
-		    {
-			type->tt_argcount -= tv->vval.v_partial->pt_argc;
-			type->tt_min_argcount -= tv->vval.v_partial->pt_argc;
-			if (type->tt_argcount > 0
-				&& func_type_add_arg_types(type,
-					    type->tt_argcount, type_gap) == OK)
-			    for (int i = 0; i < type->tt_argcount; ++i)
-				type->tt_args[i] =
-					ufunc->uf_func_type->tt_args[
-					      i + tv->vval.v_partial->pt_argc];
-		    }
-		    return type;
-		}
+		if (tv->v_type == VAR_PARTIAL && tv->vval.v_partial != NULL)
+		    return partial_typval2type(tv, ufunc, type_gap);
 		return ufunc->uf_func_type;
 	    }
 	}
@@ -719,7 +730,7 @@ check_typval_type(type_T *expected, typval_T *actual_tv, where_T where)
 
     if (expected == NULL)
 	return OK;  // didn't expect anything.
-		    //
+
     ga_init2(&type_list, sizeof(type_T *), 10);
 
     // A null_function and null_partial are special cases, they can be used to
@@ -737,12 +748,14 @@ check_typval_type(type_T *expected, typval_T *actual_tv, where_T where)
     {
 	res = check_type_maybe(expected, actual_type, TRUE, where);
 	if (res == MAYBE && !(actual_type->tt_type == VAR_FUNC
-				      && actual_type->tt_member == &t_unknown))
+				      && (actual_type->tt_member == &t_unknown
+					  || actual_type->tt_member == NULL)))
 	{
 	    // If a type check is needed that means assigning "any" or
 	    // "unknown" to a more specific type, which fails here.
 	    // Except when it looks like a lambda, since they have an
-	    // incomplete type.
+	    // incomplete type.  A legacy lambda function has a NULL return
+	    // type.
 	    type_mismatch_where(expected, actual_type, where);
 	    res = FAIL;
 	}
@@ -1739,8 +1752,15 @@ type_name(type_T *type, char **tofree)
 
     if (type->tt_type == VAR_OBJECT || type->tt_type == VAR_CLASS)
     {
-	char_u *class_name = type->tt_class == NULL ? (char_u *)"Unknown"
-				    : type->tt_class->class_name;
+	char_u *class_name;
+	if (type->tt_class != NULL)
+	{
+	    class_name = type->tt_class->class_name;
+	    if (IS_ENUM(type->tt_class))
+		name = "enum";
+	}
+	else
+	    class_name = (char_u *)"Unknown";
 	size_t len = STRLEN(name) + STRLEN(class_name) + 3;
 	*tofree = alloc(len);
 	if (*tofree != NULL)
@@ -1867,15 +1887,28 @@ f_typename(typval_T *argvars, typval_T *rettv)
     int
 check_typval_is_value(typval_T *tv)
 {
-    if (tv->v_type == VAR_CLASS)
+    if (tv == NULL)
+	return OK;
+
+    switch (tv->v_type)
     {
-        semsg(_(e_using_class_as_value_str), tv->vval.v_class->class_name);
-	return FAIL;
-    }
-    else if (tv->v_type == VAR_TYPEALIAS)
-    {
-        semsg(_(e_using_typealias_as_value_str), tv->vval.v_typealias->ta_name);
-	return FAIL;
+	case VAR_CLASS:
+	    {
+		class_T *cl = tv->vval.v_class;
+		if (IS_ENUM(cl))
+		    semsg(_(e_using_enum_as_value_str), cl->class_name);
+		else
+		    semsg(_(e_using_class_as_value_str), cl->class_name);
+	    }
+	    return FAIL;
+
+	case VAR_TYPEALIAS:
+	    semsg(_(e_using_typealias_as_value_str),
+		    tv->vval.v_typealias->ta_name);
+	    return FAIL;
+
+	default:
+	    break;
     }
     return OK;
 }
@@ -1886,36 +1919,27 @@ check_typval_is_value(typval_T *tv)
     int
 check_type_is_value(type_T *type)
 {
-    if (type->tt_type == VAR_CLASS)
+    if (type == NULL)
+	return OK;
+    switch (type->tt_type)
     {
-        semsg(_(e_using_class_as_value_str), type->tt_class->class_name);
-	return FAIL;
-    }
-    else if (type->tt_type == VAR_TYPEALIAS)
-    {
-	// Not sure what could be done here to get a name
-	// TODO: MAYBE AN OPTIONAL ARGUMENT
-        emsg(_(e_using_typealias_as_var_val));
-	return FAIL;
-    }
-    return OK;
-}
+	case VAR_CLASS:
+	    if (IS_ENUM(type->tt_class))
+		semsg(_(e_using_enum_as_value_str),
+			type->tt_class->class_name);
+	    else
+		semsg(_(e_using_class_as_value_str),
+			type->tt_class->class_name);
+	    return FAIL;
 
-/*
- * Same as above, except check vartype_T.
- */
-    int
-check_vartype_is_value(vartype_T typ)
-{
-    if (typ == VAR_CLASS)
-    {
-	emsg(_(e_using_class_as_var_val));
-	return FAIL;
-    }
-    else if (typ == VAR_TYPEALIAS)
-    {
-        emsg(_(e_using_typealias_as_var_val));
-        return FAIL;
+	case VAR_TYPEALIAS:
+	    // TODO: Not sure what could be done here to get a name.
+	    //       Maybe an optional argument?
+	    emsg(_(e_using_typealias_as_var_val));
+	    return FAIL;
+
+	default:
+	    break;
     }
     return OK;
 }
